from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.models import Prompt
from nextline_rdb.models.strategies import (
    st_model_prompt,
    st_model_prompt_list,
    st_model_run,
    st_model_trace_list,
)
from nextline_rdb.utils.strategies import st_none_or

from ...db import AsyncDB


@given(st.data())
async def test_st_model_prompt(data: st.DataObject) -> None:
    prompt = data.draw(st_model_prompt())

    async with AsyncDB() as db:
        async with db.session.begin() as session:
            session.add(prompt)
        async with db.session() as session:
            select_prompt = select(Prompt)
            prompt_ = await session.scalar(
                select_prompt.options(
                    selectinload(Prompt.run), selectinload(Prompt.trace)
                )
            )
            assert prompt_
            session.expunge_all()
    assert repr(prompt) == repr(prompt_)


@given(st.data())
async def test_st_model_prompt_lists(data: st.DataObject) -> None:
    # generate_traces=False because that would generate a trace with prompts
    if run := data.draw(st_none_or(st_model_run(generate_traces=False))):
        data.draw(st_none_or(st_model_trace_list(run=run, min_size=0, max_size=5)))
        assert run.traces is not None

    max_size = data.draw(st.integers(min_value=0, max_value=10))

    prompts = data.draw(st_model_prompt_list(run=run, max_size=max_size))

    assert len(prompts) <= max_size

    if prompts:
        runs = set(prompt.trace.run for prompt in prompts)
        assert len(runs) == 1
        assert run is None or run is runs.pop()

    async with AsyncDB() as db:
        async with db.session.begin() as session:
            session.add_all(prompts)

        async with db.session() as session:
            select_prompt = select(Prompt)
            prompts_ = (await session.scalars(select_prompt)).all()
            session.expunge_all()
    prompts = sorted(prompts, key=lambda prompt: prompt.id)
    assert repr(prompts) == repr(sorted(prompts_, key=lambda prompt: prompt.id))



from hypothesis import given, settings
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.models import Prompt, Run, Stdout, Trace
from nextline_rdb.models.strategies import st_model_run, st_model_run_list
from nextline_rdb.utils import safe_compare as sc
from nextline_rdb.utils.strategies import (
    st_datetimes,
    st_graphql_ints,
    st_none_or,
    st_ranges,
)

from ...db import AsyncDB


@given(st.data())
@settings(max_examples=200)
async def test_st_model_run(data: st.DataObject) -> None:
    run_no = data.draw(st_none_or(st_graphql_ints(min_value=1)))
    if run_no is None:
        min_run_no, max_run_no = data.draw(st_ranges(st_=st_graphql_ints, min_start=1))
    else:
        min_run_no, max_run_no = None, None

    min_started_at, max_started_at = data.draw(st_ranges(st_=st_datetimes))
    min_ended_at, max_ended_at = data.draw(
        st_ranges(st_=st_datetimes, min_start=min_started_at)
    )

    generate_traces = data.draw(st.booleans())

    run = data.draw(
        st_model_run(
            run_no=run_no,
            min_run_no=min_run_no,
            max_run_no=max_run_no,
            min_started_at=min_started_at,
            max_started_at=max_started_at,
            min_ended_at=min_ended_at,
            max_ended_at=max_ended_at,
            generate_traces=generate_traces,
        )
    )

    if run_no is not None:
        assert run.run_no == run_no
    else:
        assert sc(min_run_no) <= run.run_no <= sc(max_run_no)

    assert sc(min_started_at) <= sc(run.started_at) <= sc(max_started_at)
    assert sc(min_ended_at) <= sc(run.ended_at) <= sc(max_ended_at)
    assert sc(run.started_at) <= sc(run.ended_at)

    if not run.started_at:
        assert not run.ended_at

    traces = run.traces
    prompts = run.prompts
    stdouts = run.stdouts
    if generate_traces:
        assert traces
    else:
        assert not traces
        assert not prompts
        assert not stdouts

    async with AsyncDB() as db:
        async with db.session.begin() as session:
            session.add(run)
        async with db.session() as session:
            select_run = select(Run)
            run_ = await session.scalar(
                select_run.options(selectinload(Run.traces), selectinload(Run.prompts))
            )
            assert run_
            session.expunge_all()

    assert repr(run) == repr(run_)
    traces = sorted(traces, key=lambda trace: trace.id)
    assert repr(traces) == repr(sorted(run_.traces, key=lambda trace: trace.id))
    prompts = sorted(prompts, key=lambda prompt: prompt.id)
    assert repr(prompts) == repr(sorted(run_.prompts, key=lambda prompt: prompt.id))
    stdouts = sorted(stdouts, key=lambda stdout: stdout.id)


@given(st.data())
@settings(max_examples=200)
async def test_st_model_run_lists(data: st.DataObject) -> None:
    max_size = data.draw(st.integers(min_value=0, max_value=10))
    runs = data.draw(st_model_run_list(generate_traces=True, max_size=max_size))
    assert len(runs) <= max_size
    run_nos = [run.run_no for run in runs]
    assert run_nos == sorted(run_nos)

    traces = [trace for run in runs for trace in run.traces]
    prompts = [prompt for run in runs for prompt in run.prompts]
    stdouts = [stdout for run in runs for stdout in run.stdouts] 

    started_ats = [run.started_at for run in runs if run.started_at]
    assert started_ats == sorted(started_ats)

    async with AsyncDB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)

        async with db.session() as session:
            select_run = select(Run)
            runs_ = (await session.scalars(select_run)).all()
            select_trace = select(Trace)
            traces_ = (await session.scalars(select_trace)).all()
            select_prompt = select(Prompt)
            prompts_ = (await session.scalars(select_prompt)).all()
            select_stdout = select(Stdout)
            stdouts_ = (await session.scalars(select_stdout)).all()
            session.expunge_all()

    assert repr(runs) == repr(runs_)
    traces = sorted(traces, key=lambda trace: trace.id)
    assert repr(traces) == repr(sorted(traces_, key=lambda trace: trace.id))
    prompts = sorted(prompts, key=lambda prompt: prompt.id)
    assert repr(prompts) == repr(sorted(prompts_, key=lambda prompt: prompt.id))
    stdouts = sorted(stdouts, key=lambda stdout: stdout.id)
    assert repr(stdouts) == repr(sorted(stdouts_, key=lambda stdout: stdout.id))



import datetime

from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.models import Stdout
from nextline_rdb.models.strategies import st_model_stdout

from ..db import AsyncDB


@given(st.data())
async def test_repr(data: st.DataObject):
    async with AsyncDB() as db:
        async with db.session.begin() as session:
            model = data.draw(st_model_stdout())
            session.add(model)

        async with db.session() as session:
            rows = await session.scalars(
                select(Stdout).options(
                    selectinload(Stdout.run), selectinload(Stdout.trace)
                )
            )
            for row in rows:
                repr_ = repr(row)
                assert Stdout, datetime  # type: ignore[truthy-function]
                assert repr_ == repr(eval(repr_))



from collections.abc import AsyncIterator, Mapping, MutableMapping
from logging import getLogger
from pathlib import Path
from typing import Optional

from apluggy import asynccontextmanager
from dynaconf import Dynaconf, Validator
from nextline import Nextline
from nextlinegraphql.hook import spec
from sqlalchemy import func, select
from sqlalchemy.orm import Session

from . import models
from .db import DB
from .schema import Mutation, Query, Subscription
from .write import write_db

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (Validator("DB.URL", must_exist=True, is_type_of=str),)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[tuple[str, ...]]:
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf) -> None:
        self._url = settings.db['url']

    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    @asynccontextmanager
    async def lifespan(self, context: Mapping) -> AsyncIterator[None]:
        nextline = context['nextline']
        self._db = DB(self._url)
        with self._db:
            await self._initialize_nextline(nextline)
            async with write_db(nextline, self._db):
                yield

    async def _initialize_nextline(self, nextline: Nextline) -> None:
        run_no, script = self._last_run_no_and_script()
        if run_no is not None:
            run_no += 1
            if run_no >= nextline._init_options.run_no_start_from:
                nextline._init_options.run_no_start_from = run_no
        if script is not None:
            nextline._init_options.statement = script

    def _last_run_no_and_script(self) -> tuple[Optional[int], Optional[str]]:
        with self._db.session() as session:
            last_run = self._last_run(session)
            if last_run is None:
                return None, None
            else:
                return last_run.run_no, last_run.script

    def _last_run(self, session: Session) -> Optional[models.Run]:
        stmt = select(models.Run, func.max(models.Run.run_no))
        if model := session.execute(stmt).scalar_one_or_none():
            return model
        else:
            logger = getLogger(__name__)
            msg = "No previous runs were found in the DB"
            logger.info(msg)
            return None

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['db'] = self._db



from logging import getLogger
from pathlib import Path
from typing import Optional

from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import sessionmaker

import nextline_rdb
from nextline_rdb.utils import ensure_sync_url

from . import models

ALEMBIC_INI = str(Path(nextline_rdb.__file__).resolve().parent / 'alembic.ini')


assert Path(ALEMBIC_INI).is_file()


def create_tables(engine: Engine) -> None:
    '''Define tables in the database based on the ORM models .

    https://docs.sqlalchemy.org/en/20/orm/quickstart.html#emit-create-table-ddl
    '''
    models.Model.metadata.create_all(bind=engine)


class DB:
    '''The interface to the SQLAlchemy database.

    An example usage:
    >>> with DB() as db:
    ...     # Nested session contexts
    ...     with db.session() as session:
    ...         with session.begin():
    ...             pass
    ...     # Direct begin
    ...     with db.session.begin() as session:
    ...         pass

    An alternative usage:
    >>> db = DB()
    >>> db.start()
    >>> with db.session() as session:
    ...     with session.begin():
    ...         pass
    >>> db.close()

    '''

    def __init__(
        self,
        url: Optional[str] = None,
        create_engine_kwargs: Optional[dict] = None,
    ):
        url = url or 'sqlite://'
        self.url = ensure_sync_url(url)
        self.create_engine_kwargs = create_engine_kwargs or {}
        self.engine = create_engine(self.url, **self.create_engine_kwargs)
        self.migration_revision: str | None = None

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.url!r}>'

    def start(self) -> None:
        logger = getLogger(__name__)
        logger.info(f"SQLAlchemy DB URL: {self.url}")
        migrate_to_head(self.engine)
        create_tables(self.engine)  # NOTE: unnecessary as alembic is used
        with self.engine.connect() as connection:
            context = MigrationContext.configure(connection)
            self.migration_revision = context.get_current_revision()
        logger.info(f"Alembic migration version: {self.migration_revision!s}")
        self.session = sessionmaker(self.engine, expire_on_commit=False)

    def close(self) -> None:
        pass

    def __enter__(self) -> 'DB':
        self.start()
        return self

    def __exit__(self, *_, **__) -> None:
        self.close()


def migrate_to_head(engine: Engine) -> None:
    '''Run alembic to upgrade the database to the latest version.'''
    config = Config(ALEMBIC_INI)

    # config.set_main_option('sqlalchemy.url', str(engine.url))
    # from alembic import command
    # command.upgrade(config, "head")

    # NOTE: The commented out lines of command.upgrade() above would work fine
    # for a persistent DB. Here, the following code is instead used so that the
    # migration actually takes place for an in-memory DB as well for testing.

    script = ScriptDirectory.from_config(config)

    def upgrade(rev: str, context):  # type: ignore[no-untyped-def]
        del context
        return script._upgrade_revs('head', rev)

    context = EnvironmentContext(config, script, fn=upgrade)
    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=models.Model.metadata,
            render_as_batch=True,
        )
        with context.begin_transaction():
            context.run_migrations()



from async_asgi_testclient import TestClient
from nextlinegraphql.plugins.ctrl.test import run_statement
from nextlinegraphql.plugins.graphql.test import gql_request

from .schema.graphql import QUERY_HISTORY


async def test_one(client: TestClient):
    await run_statement(client)
    data = await gql_request(client, QUERY_HISTORY)
    runs = data['history']['runs']
    edges = runs['edges']
    assert 2 == len(edges)
    run = edges[1]['node']
    assert 2 == run['runNo']
    assert 'finished' == run['state']
    assert run['startedAt']
    assert run['endedAt']
    assert run['script']
    assert not run['exception']



from collections.abc import Callable

import pytest
from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import inspect

from nextline_rdb.db import DB
from nextline_rdb.models import Model
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.utils import ensure_async_url


def object_state(obj: Model) -> str:
    insp = inspect(obj)
    if insp.transient:
        return 'transient'
    elif insp.pending:
        return 'pending'
    elif insp.persistent:
        return 'persistent'
    elif insp.deleted:
        return 'deleted'
    elif insp.detached:
        return 'detached'
    else:
        raise ValueError(f'Unknown state for {obj}')


def test_ensure_sync_url(tmp_url_factory: Callable[[], str]):
    url = tmp_url_factory()
    async_url = ensure_async_url(url)

    with DB(url=async_url) as db:
        assert db.url == url


def test_migration_revision() -> None:
    with DB() as db:
        assert db.migration_revision


@given(st.data())
def test_session_nested(tmp_url_factory: Callable[[], str], data: st.DataObject):
    runs = data.draw(st_model_run_list(generate_traces=True, max_size=10))

    url = tmp_url_factory()

    with DB(url) as db:
        with db.session() as session:
            with session.begin():
                session.add_all(runs)
                saved = sorted(session.new, key=lambda x: (x.__class__.__name__, x.id))
                model_classes = {type(x) for x in saved}
        repr_saved = [repr(m) for m in saved]

        with db.session() as session:
            loaded = sorted(
                (m for cls in model_classes for m in session.query(cls)),
                key=lambda x: (x.__class__.__name__, x.id),
            )
            repr_loaded = [repr(m) for m in loaded]

        assert repr_saved == repr_loaded


@given(st.data())
def test_session_begin(tmp_url_factory: Callable[[], str], data: st.DataObject):
    runs = data.draw(st_model_run_list(generate_traces=True, max_size=10))

    url = tmp_url_factory()

    with DB(url) as db:
        with db.session.begin() as session:
            session.add_all(runs)
            saved = sorted(session.new, key=lambda x: (x.__class__.__name__, x.id))
            model_classes = {type(x) for x in saved}
        repr_saved = [repr(m) for m in saved]

        with db.session() as session:
            loaded = sorted(
                (m for cls in model_classes for m in session.query(cls)),
                key=lambda x: (x.__class__.__name__, x.id),
            )
            repr_loaded = [repr(m) for m in loaded]

        assert repr_saved == repr_loaded


@pytest.fixture(scope='session')
def tmp_url_factory(tmp_path_factory: pytest.TempPathFactory) -> Callable[[], str]:
    def factory() -> str:
        dir = tmp_path_factory.mktemp('db')
        url = f'sqlite:///{dir}/db.sqlite'
        return url

    return factory



from typing import Optional, Type

from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, sessionmaker

from nextline_rdb import models
from nextline_rdb.utils import ensure_async_url, ensure_sync_url


class DB:
    '''The interface to the SQLAlchemy database.

    >>> db = DB()
    >>> with db.session() as session:
    ...     with session.begin():
    ...         pass

    >>> with db.session.begin() as session:
    ...     pass

    '''

    def __init__(
        self,
        url: Optional[str] = None,
        model_base_class: Type[DeclarativeBase] = models.Model,
        echo: bool = False,
    ):
        url = url or 'sqlite://'
        self.url = ensure_sync_url(url)
        self.model_base_class = model_base_class
        self.metadata = self.model_base_class.metadata
        self.engine = create_engine(self.url, echo=echo)
        self.metadata.create_all(bind=self.engine)
        self.session = sessionmaker(self.engine, expire_on_commit=False)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.url!r}>'


class AsyncDB:
    '''The interface to the async SQLAlchemy database.

    >>> async def main():
    ...     # An example usage
    ...     async with AsyncDB() as db:
    ...         # nested session contexts
    ...         async with db.session() as session:
    ...             async with session.begin():
    ...                 # commit automatically
    ...                 pass
    ...         # can begin directly
    ...         async with db.session.begin() as session:
    ...             # commit automatically
    ...             pass
    ...
    ...     # An alternative usage
    ...     db = AsyncDB()
    ...     await db.start()
    ...     async with contextlib.aclosing(db):
    ...         async with db.session() as session:
    ...             async with session.begin():
    ...                 pass
    ...
    >>> import asyncio
    >>> import contextlib
    >>> asyncio.run(main())

    '''

    def __init__(
        self,
        url: Optional[str] = None,
        model_base_class: Type[DeclarativeBase] = models.Model,
        echo: bool = False,
    ):
        url = url or 'sqlite+aiosqlite://'
        self.url = ensure_async_url(url)
        self.model_base_class = model_base_class
        self.metadata = self.model_base_class.metadata
        self.engine = create_async_engine(self.url, echo=echo)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.url!r}>'

    async def start(self) -> None:
        async with self.engine.begin() as conn:
            await conn.run_sync(self.metadata.create_all)
        self.session = async_sessionmaker(self.engine, expire_on_commit=False)

    async def aclose(self) -> None:
        await self.engine.dispose()

    async def __aenter__(self) -> 'AsyncDB':
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:  # type: ignore[no-untyped-def]
        del exc_type, exc_value, traceback
        await self.aclose()



import contextlib

import pytest

from .db import AsyncDB


@pytest.fixture
def db():
    return AsyncDB()


async def test_fields(db: AsyncDB):
    assert db.url
    assert db.metadata
    assert db.engine
    await db.start()
    async with contextlib.aclosing(db):
        assert db.session


async def test_session(db: AsyncDB):
    async with db:
        async with db.session() as session:
            async with session.begin():
                pass


async def test_session_maker(db: AsyncDB):
    async with db:
        async with db.session.begin() as session:
            assert session
            pass



import pytest

from .db import DB


@pytest.fixture
def db():
    return DB()


def test_fields(db: DB):
    assert db.url
    assert db.metadata
    assert db.engine
    assert db.session


def test_session(db: DB):
    with db.session() as session:
        with session.begin():
            pass


def test_session_maker(db: DB):
    with db.session.begin() as session:
        assert session
        pass



__all__ = [
    'Model',
    'repr_val',
    'Hello',
    'Run',
    'Trace',
    'Prompt',
    'Stdout',
]


from .base import Model, repr_val
from .hello import Hello
from .prompt import Prompt
from .run import Run
from .stdout import Stdout
from .trace import Trace
